/*
  Macro           CursorLine
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Compatibility   Windows TSE GUI v4.2 upwards.
                  Linux TSE v4.50 rc 17 upwards, needing RGB support from Linux.
  Version         v1.3   13 Jan 2024

  This extension gives you 3 options for TSE's cursor line:
  - Color the cursor line the same as other lines.
  - Use TSE's cursor line colors.
  - Dim or brighten the cursor line's foreground and background colors
    by a configurable percentage.

  Switch between options retains the other's settings.


  LINUX REQUIREMENTS

  Linux TSE v4.50 rc 17 upwards supports RGB colors, but your Linux terminal
  or command line might not.
  Most modern ones do, but there are still exceptions.
  TSE cannot check if yours does.
  TSE can only blindly send ANSI RGB codes to the screen.
  TSE will start sending ANSI RGB codes to the screen:
  - If TSE is started with "e --ansicolor 1".
  - After setting the configuration menu "ANSI Color driver" to "RGB",
    saving settings,
    restarting TSE.
  You will visually have to determine if this works or not.

  If it does NOT work then TSE might leave the screen mangled.

  Probable solutions to a mangled screen:
  - Restart the computer.
  - One of these Linux commands:
    - printf "\ec"    (erases screen scroll history)
    - reset

  A way to avoid a mangled screen is to start Linux TSE with
    e -ansicolor 8
  and to test the Linux terminal/command line's RGB compatibility
  with v1.1 or later of the Test_RGB_color_support tool:
    https://ecarlo.nl/tse/DemosAndTests.html#Test_RGB_color_support


  INSTALLATION

  Copy this file to TSE's "mac" folder and compile it there, for example by
  opening it there in TSE and applying the Macro, Compile menu.

  This extension needs to be configured for it to do something.


  CONFIGURATION

  Execute the extension as a macro, for example by selecting the
  menu Macro -> Execute, and in the prompt enter "CursorLine".

  Do not add this extension to TSE's Macro AutoLoad List yourself.
  It adds and deletes itself based on your configuration choices.


  DIM AND BRIGHTEN

  Default a percentage is 0%, meaning colors stay the same.
  The brightest percentage is  100%, making all impacted colors white.
  The dimmest   percentage is -100%, making all impacted colors black.


  VIDEO

  There is a demo CursurLine video at my website:
    https://ecarlo.nl/tse/ToolsAndExtensions.html#CursorLine


  TIPS

  In general TSE works best if it is configured to either have dark letter
  colors on bright background colors or to have bright letter colors on dark
  background colors.

  Both this extension and TSE's own "Syntax Hilite inside Blocks" option work
  best, if the above configuration choice is the same for the text and block
  colors: Either both bright on dark or both dark on bright.

  Because letters take up less space than their background:
  When brightening or dimming the cursor line's colors with this extension,
  its foreground (letter) colors need to change by a large percentage for the
  change to be visible, while its background colors only need to change by a
  small percentage for the change to be visible.
  Note that dimming and brightening foreground (letter) colors will make their
  syntax hiliting vaguer.
  In other words, dimming or brightening the background colors works better.


  NERDY MENU POINTS

  I nerded out on making the configuration menu user-friendlier than TSE's
  standard menus allow:
    1. The menu has a close button.
    2. The menu extends down for its submenus.                      (*)
    3. Two menu options split their label and value over two lines. (*)
    Two menu options let a user decrease and increase their value with:
    4. Pressing the   left and right cursor keys.
    5. Clicking their left and right arrow buttons.

  (*)
    This was done to keep the menu slim, anticipating that someday a preview
    pane might be added on the right side of the screen.


  KNOWN IMPERFECTIONS

  If TSE's vertical screen size is less than 29 rows, then selecting a TSE
  color from the second submenu opens a window that overlaps the menu.
  This is not a bug. It is just less pretty and a tad less user-friendly,
  because it obscures which of the two colors is being selected.

  If the cursor line coloring type is "Same as other lines" and you change a
  cursor line color outside this extension, then the extension might not become
  aware of this until its configuration is started.


  TODO
    MUST
    SHOULD
    COULD
    - When configuring, show a preview pane on the right side of the screen.
      This whish seems technically incompatible with the configuration's use
      of a TSE menu.
      The problem is that the macro language does not support the use of
      concurrent video outpout windows.
      Possible solutions:
      - Write my own non-TSE menu using low-level macro commands.
        That would be excessive work for me.
      - Ask Semware for macro syntax to support concurrent video outout windows.
        That would be excessive work for Semware.
      - Ask Semware for a PopWinClose() parameter that makes it not restore its
        windows's video output.
        From my outside point of view this seems a reasonable solution.
        I will wait asking Semware for this at least until I have a more
        finished version of this macro to point to as a use-case.
        And even then I might not: I think it would be nice to have and would
        theoretically open up new possibilities, but currently I do not have
        an urgent practical need for it.
    - Make the dim/brighten settings settable per color scheme.
      I can think of theoretical examples where that might seem handy, but my
      guestimate is that those examples would be sparsely used in practice, if
      at all.
    WONT
    - The below ideas are technically possible, but might perform less well and
      would be time-consuming to implement. While they would look amazing, I
      would probably not use them in practice. So my cost/benefit analysis says
      not to do them:
      - Add a few less dimmed/brightened lines around the cursor line,
        proportionally to their distance from it.
      - Put a round spotlight on the cursor.


  HISTORY

  v1       13 Oct 2021
    Initial release.

  v1.0.1   13 Oct 2021
    Fixed it not actually being backwards compatible
    with Windows TSE Pro GUI v4.2.

  v1.0.2   30 Nov 2021
    Fixed the lag in brightening and dimming the cursor line.
    Improved the documentation.

  v1.0.3   17 Sep 2022
    Fixed incompatibility with TSE's '-i' command line option
    and the TSELOADDIR environment variable.

  v1.0.4   14 Oct 2022
    Fixed a future problem: CursurLine now disables itself during _DISPLAY_HELP_
    display mode, so it no longer overwrites OEM characters with ANSI characters
    there. Currently nobody edits files in _DISPLAY_HELP_ display mode, but in
    the future some of us will.

  v1.1      3 Sep 2023
    CursorLine can now be used with g32.exe and e32.exe in the same directory.
    In such a case g32.exe and e32.exe share the same autoloaded files.
    CursorLine is still incompatible with e32.exe, but when e32.exe autoloads
    it, its error message now asks whether to repeat the message in the future.
    So default Console users still get the error message, but they can now
    disable it.

  v1.2     25 Sep 2023
    Major but backwards-compatible update:
    - Now supports 3 cursor line options: Same, TSE, and Dim/brighten.
    - Now switching between them does not need you to turn TSE's cursor line
      coloring on/off yourself: It does so automatically and remembers old
      values when you switch back.
    - Now has a much user-friendlier configuration menu.

  v1.3     13 Jan 2024
    Made the extension Linux-compatible.

*/





/*
  T E C H I C A L   I N F O   F O R   M A C R O   P R O G R A M M E R S


  TERMINOLOGY

  When TSE functions refer to a "color", with 2 exceptions they actually refer
  to a "color attribute".
  Such a color attribute has an 8-bit value:
  The upper 4 bits are an index into a table of 16 actual background colors.
  The lower 4 bits are an index into a table of 16 actual foreground colors.
  In an RGB color-capable environment these actual colors have 24-bit values.
  This means an actual color can be one of 16,777,216 possible values.
  The actual colors for these two 16-value tables can at any moment be queried
  and set by the GetColorTableValue() and SetColorTableValue() functions.
  These two tables, the one for the 16 foreground colors and the one for the 16
  background colors, can hold different sets of actual colors.


  TRICK

  Oversimplified:

  The editor itself regularly updates the screen.

  If a macro temporarily assigns a new RGB color to one or both of the two
  color tables and updates the screen with PutStrAttrXY(), then this additional
  RGB color is shown for as long as the editor itself does not update the
  screen.

  In theory all 16,777,216 RGB colors can simultenously be shown this way.
  In practice that number is limited to the number of characters that can
  simultaneously be shown on the screen.

  The above method is anything but simple. It is fraught with obstacles.
  Below these obstacles are described.


  THE PUTSTRATTRXY() BUG

  The bug for TSE versions up to and including TSE 4.2 is,
  that if PutStrAttrXY() is directed to write to the screen with color
  attribute 0 (BLACK on BLACK), then it uses color attribute Query(Attr)
  instead.
  This can relatively easy be worked around by pre-setting Attr to 0 before
  using PutStrAttrXY() to write color 0.


  WORKING AROUND SCREEN CACHING

  This is a huge problem.

  The following kept mentally tripping me up:
  The TSE documentation refers to a character attribute's "color".
  There are even predefined constants that give names to "colors".
  But actually these references are to a color attribute, which is two indexes,
  one into a table of 16 actual background colors, and one into a table of 16
  actual foreground colors.
  Up to TSE v4.2 that difference did not matter, because there was only one
  fixed set of actual colors.
  So it did made sense to simplify the documentation this way.
  However, with the introduction of the GetColorTableValue() and
  SetColorTableValue() functions in TSE v4.2 that no longer makes sense.
  The documentation was never updated, except for those two functions.
  So we have to mentally adjust for the outdated documentation.
  I kept being tripped up by confusing color attributes with the actual colors
  they refer to.

  Now to TSE's color ATTRIBUTE caching:
  The PutStrAttrXY() function does not recolor the screen when it writes the
  same color attribute that is already there.
  Normally this is a good thing, because it optimizes TSE's screen updates.
  Here however, it prohibites us from using PutStrAttrXY() with the same color
  attribute after having used SetColorTableValue() to temporarily set a new RGB
  color for the existing color attribute part.
  So to write a new RGB color to the screen we need to use a new color
  attribute.
  But a color attribute consists of a 4-bit reference to a background color and
  a 4-bit reference to a foreground color.
  So using a new color attribute would potentially reference 2 new actual
  colors!

  A general solution is possible, but is from complex to very complex.
  Firstly, it requires updating the screen twice; once to an unused dummy
  attribute to force a dummy screen update and once to the intended new or old
  attribute to force the needed screen update.
  Secondly, if the application needs more than 16 background/foreground colors
  on the same line, then it requires updating the screen character by character,
  which at best is visibly slow.

  Luckily this extension has simpler requirements that can use a simpler
  solution.
  Because it updates the old set of colors the same for a whole line, it can
  overwrite the line as a whole, twice, which is fast.
  It first overwrites the line with the color attribute for Black on Black to
  undo TSE's screen caching, and then immediately overwrites the line with its
  old set of color attributes, which temporarily have been set to make use of
  dimmed and/or brightend RGB colors.

*/





// Start of compatibility restrictions and mitigations



/*
  When compiled for Linux or with a TSE version below TSE v4.2 the compiler
  reports a syntax error and hilights the first applicable line below.
*/

#ifndef INTERNAL_VERSION
  #define INTERNAL_VERSION 0
#endif

#define LINUX_MINIMUM_INTERNAL_VERSION 12379

#ifdef LINUX
  #define WIN32 FALSE
  #if INTERNAL_VERSION < LINUX_MINIMUM_INTERNAL_VERSION



    for LINUX this extension needs at least TSE v4.50 release candidate 17.



  #endif
#endif

#ifndef WIN32
   16-bit versions of TSE are not supported. You need at least TSE Pro v4.2.
#endif

#ifndef EDITOR_VERSION
   Editor Version is older than TSE Pro v3.0. You need at least TSE Pro v4.2.
#endif

#if EDITOR_VERSION < 4200h
   Editor Version is older than TSE Pro v4.2. You need at least TSE Pro v4.2.
#endif



/*
  isAutoLoaded1() v1.0

  Syntax:
    integer isAutoLoaded1(string macro_name)

  You may supply a fully qualified macro name, like CurrMacroFilename()
  returns. IsAutoLoaded1() will only use the name part.
  TSE's IsAutoLoaded() does not do this: It is an added feature
  of IsAutoLoaded1().

  For TSE 4.4 upwards IsAutoLoaded1() will just use TSE's IsAutoloaded
  function (with the split-off macro_name).

  For older TSE versions IsAutoLoaded1() implements TSE 4.4's isAutoLoaded()
  function *with* a parameter.

  So, for IsAutoLoaded1() the macro_name parameter is mandatory.
  An empty string as a parameter will return FALSE, just like TSE 4.4's
  IsAutoLoaded('') does.
*/

#if EDITOR_VERSION < 4400h
  integer isAutoLoaded1_id                    = 0
  string  isAutoLoaded1_file_name_chrset [44] = "- !#$%&'()+,.0-9;=@A-Z[]^_`a-z{}~\d127-\d255"
#endif

integer proc isAutoLoaded1(string macro_name_in)
  string  macro_name    [255] = SplitPath(macro_name_in, _NAME_)
  integer result              = FALSE
  #if EDITOR_VERSION >= 4400h
    result = isAutoLoaded(macro_name)
  #else
    string  autoload_name [255] = ''
    string  old_wordset    [32] = Set(WordSet, ChrSet(isAutoLoaded1_file_name_chrset))
    integer org_id              = GetBufferId()

    if isAutoLoaded1_id
      GotoBufferId(isAutoLoaded1_id)
    else
      autoload_name    = SplitPath(CurrMacroFilename(), _NAME_) + ':isAutoLoaded'
      isAutoLoaded1_id = GetBufferId(autoload_name)
      if isAutoLoaded1_id
        GotoBufferId(isAutoLoaded1_id)
      else
        isAutoLoaded1_id = CreateTempBuffer()
        ChangeCurrFilename(autoload_name, _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
      endif
    endif
    LoadBuffer(LoadDir() + 'tseload.dat')
    result = lFind(macro_name, 'giw')
    Set(WordSet, old_wordset)
    GotoBufferId(org_id)
  #endif
  return(result)
end isAutoLoaded1



// End of compatibility restrictions and mitigations.





// Global constants and semi-constants
#define BACKGROUND_MENU_ROW         6
#define BLACK_ON_BLACK              Color(Black ON Black)
#define CCF_OPTIONS                 _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_
string  COLOR_NAMES          [46] = 'Black Blue Green Cyan Red Magenta Yellow White'
#define CURSOR_ATTR                 1
#define CURSOR_IN_BLOCK_ATTR        2
#define CURSOR_LINE_ROW             5
#define CURSOR_LINE_IN_BLOCK_ROW    CURSOR_LINE_ROW + 2
#define FOREGROUND_MENU_ROW         BACKGROUND_MENU_ROW + 1
integer HASHES_ID                 = 0
string  MACRO_NAME [MAXSTRINGLEN] = ''
#define NO_POSITIONS_NO_SHADOWS     _USE_3D_CHARS_|_USE_3D_BUTTONS_



// Global variables
integer background_percentage           = 0
integer brightened_buffer_id            = 0
integer brightened_file_line_nr         = 0
integer brightened_screen_row_nr        = 0
integer brightened_window_row_nr        = 0
integer color_type                      = 0
integer foreground_percentage           = 0
integer main_menu_history               = 1
string  saved_rgb_colors [MAXSTRINGLEN] = ''
integer update_cursor_line              = -1



forward proc after_getkey()


integer proc tse_has_rgb_support()
  integer result = FALSE
  #ifdef LINUX
    #if INTERNAL_VERSION >= LINUX_MINIMUM_INTERNAL_VERSION
      if Query(CurrAnsiColor) == 1  // 1 == RGB.
        result = TRUE
      endif
    #endif
  #else
    if isGUI()
      result = TRUE
    endif
  #endif
  return(result)
end tse_has_rgb_support


integer proc tse_is_windows_console_version()
  integer result = FALSE
  #ifndef LINUX
    if not isGUI()
      result = TRUE
    endif
  #endif
  return(result)
end tse_is_windows_console_version


integer proc write_profile_int(string  item_name,
                               integer item_value)
  integer result = WriteProfileInt(MACRO_NAME + ':Config',
                                   item_name,
                                   item_value)
  if not result
    Alarm()
    Warn(MACRO_NAME; 'failed to write configuration to file "tse.ini".')
  endif
  return(result)
end write_profile_int

// Abbreviation of GetProfileInt() to make the calling code more readable.
integer proc get_profile_int(string item_name)
  return(GetProfileInt(MACRO_NAME + ':Config', item_name, 0))
end get_profile_int



/*


integer proc write_profile_str(string section_name,
                               string item_name,
                               string item_value)
  integer result = 0
  result = WriteProfileStr(section_name,
                           item_name,
                           item_value)
  if not result
    Alarm()
    Warn('ERROR writing'; MACRO_NAME, "'s configuration to TSE's .ini file.")
  endif
  return(result)
end write_profile_str


proc show_help()
  string  full_macro_source_name [MAXSTRINGLEN] = SplitPath(CurrMacroFilename(), _DRIVE_|_PATH_|_NAME_) + '.s'
  string  help_file_name         [MAXSTRINGLEN] = '*** ' + MACRO_NAME + ' Help ***'
  integer hlp_id                                = GetBufferId(help_file_name)
  integer org_id                                = GetBufferId()
  integer tmp_id                                = 0
  if hlp_id
    GotoBufferId(hlp_id)
    UpdateDisplay()
  else
    tmp_id = CreateTempBuffer()
    if LoadBuffer(full_macro_source_name)
      // Separate characters, otherwise my SynCase macro gets confused.
      if lFind('/' + '*', 'g')
        PushBlock()
        UnMarkBlock()
        Right(2)
        MarkChar()
        if not lFind('*' + '/', '')
          EndFile()
        endif
        MarkChar()
        Copy()
        CreateTempBuffer()
        Paste()
        UnMarkBlock()
        PopBlock()
        BegFile()
        ChangeCurrFilename(help_file_name, CCF_OPTIONS)
        BufferType(_NORMAL_)
        FileChanged(FALSE)
        BrowseMode(TRUE)
        UpdateDisplay()
      else
        GotoBufferId(org_id)
        Warn('File "', full_macro_source_name, '" has no multi-line comment block.')
      endif
    else
      GotoBufferId(org_id)
      Warn('File "', full_macro_source_name, '" not found.')
    endif
    AbandonFile(tmp_id)
  endif
end show_help


*/



proc select_color_attribute(var integer attr_value)
  integer attr_select_value   = attr_value
  string  attr_str_value  [2] = ''
  integer background          = 0
  string  color_name     [33] = ''
  integer cols_per_color      = 2
  integer foreground          = 0
  integer inner_first_column  = 0
  integer inner_first_line    = 0
  integer inner_last_column   = 0
  integer inner_last_line     = 0
  integer left_margin         = 8
  integer mouse_background    = 0
  integer mouse_foreground    = 0
  integer old_cursor          = Set(Cursor, OFF)
//integer old_SpecialEffects  = 0
  integer outer_first_column  = 0
  integer outer_first_line    = 0
  integer outer_last_column   = 0
  integer outer_last_line     = 0
  integer prev_clockticks     = 0
  integer rows_per_color      = 1
  integer selected_background = attr_value  /  16
  integer selected_foreground = attr_value mod 16
  integer sub_row             = 0
  string  window_title  [255] = ""

  //  The config menu (above the outer window has 10 lines and 42 columns.
  //  The borderless outer window needs 19 lines.
  //  Its first line shows the color description.
  //  Its other lines are all for the inner window.
  //  The inner window shows the color table which has borders.
  //  The inner window has 16 lines for 16 colors plus 2 lines for its borders.
  if Query(ScreenRows) >= 29
    outer_first_line = 11
  else
    outer_first_line = Query(ScreenRows) - rows_per_color * 16 - 2 - 1 + 1
  endif
  outer_last_line    = outer_first_line  + rows_per_color * 16 + 2 + 1 - 1
  outer_first_column = 1
  outer_last_column  = Max(42, outer_first_column + cols_per_color + 16 + 2 - 1)

  inner_first_line   = outer_first_line   + 1
  inner_last_line    = inner_first_line   + rows_per_color * 16 + 2 - 1
  inner_first_column = outer_first_column + left_margin
  inner_last_column  = inner_first_column + cols_per_color * 16 + 2 - 1

  BufferVideo()
  Set(Attr, Color(Bright White ON Black))
//old_SpecialEffects = Set(SpecialEffects,
//                         Query(SpecialEffects) & ~ _DRAW_SHADOWS_)

  PopWinOpen(outer_first_column,
             outer_first_line,
             outer_last_column,
             outer_last_line,
             0,
             '',
             Color(Bright White ON Black))
  ClrScr()
  repeat
    color_name = iif(selected_background > 7, "Intense ", "")
               + iif(selected_foreground > 7, "Bright " , "")
               + GetToken(COLOR_NAMES, " ", (selected_foreground mod 8) + 1)
               + " on "
               + GetToken(COLOR_NAMES, " ", (selected_background mod 8) + 1)
    attr_str_value = Upper(Format(attr_select_value:2:'0':16))
    window_title   = Format('(', attr_str_value, ')', color_name:35)
    PutStrAttrXY(1, 1, window_title, "", Color(Bright White ON Black))
    ClrEol()
    PopWinOpen(inner_first_column,
               inner_first_line,
               inner_last_column,
               inner_last_line,
               1,
               '',
               Color(Bright White ON Black))
    for foreground = 0 to 15 // Rows.
      for background = 0 to 15 // Columns.
        for sub_row = 1 to rows_per_color
          if foreground <> 0 // Skip a bug in the PutStrAttrXY command
          or background <> 0 // up to and including TSE 4.2.
            PutStrAttrXY(background * cols_per_color + 1,
                         foreground * rows_per_color + sub_row,
                         Format((background * 16 + foreground)
                                :cols_per_color:"0":16),
                         "",
                         background * 16 + foreground)
          endif
        endfor
      endfor
    endfor
    PopWinOpen(inner_first_column + selected_background * cols_per_color,
               inner_first_line   + selected_foreground * rows_per_color,
               inner_first_column + selected_background * cols_per_color + cols_per_color + 1,
               inner_first_line   + selected_foreground * rows_per_color + rows_per_color + 1,
               1,
               '',
               Color(Bright White ON Black))
    UnBufferVideo()
    case GetKey()
      when <Home>, <GreyHome>
        selected_foreground = 0
        selected_background = 0
      when <end>, <GreyEnd>
        selected_foreground = 15
        selected_background = 15
      when <CursorDown>, <GreyCursorDown>
        if selected_foreground == 15
          selected_foreground = 0
        else
          selected_foreground = selected_foreground + 1
        endif
      when <CursorUp>, <GreyCursorUp>
        if selected_foreground == 0
          selected_foreground = 15
        else
          selected_foreground = selected_foreground - 1
        endif
      when <CursorRight>, <GreyCursorRight>
        if selected_background == 15
          selected_background = 0
        else
          selected_background = selected_background + 1
        endif
      when <CursorLeft>, <GreyCursorLeft>
        if selected_background == 0
          selected_background = 15
        else
          selected_background = selected_background - 1
        endif
      when <LeftBtn>
        /*
           The Delay(1) turned out to be necessary to give the system
           time to update the mouse status. An extra call to MouseStatus
           was useless, as could be expected, because TSE should already
           do that internally after a mousekey.
        */
        Delay(1)
        mouse_background = (Query(MouseX) - inner_first_column - 1) / cols_per_color
        mouse_foreground = (Query(MouseY) - inner_first_line   - 1) / rows_per_color
        if  (mouse_background in 0 .. 15)
        and (mouse_foreground in 0 .. 15)
          if  mouse_background == selected_background
          and mouse_foreground == selected_foreground
            if GetClockTicks() - prev_clockticks < 18
              PushKey(<Enter>)
            endif
          else
            selected_background = mouse_background
            selected_foreground = mouse_foreground
          endif
          prev_clockticks = GetClockTicks()
        endif
    endcase
    attr_select_value = selected_background * 16 + selected_foreground
    BufferVideo()
    PopWinClose()
    PopWinClose()
  until Query(Key) in <Enter>, <GreyEnter>, <Escape>
  UpdateDisplay(_ALL_WINDOWS_REFRESH_)
  PopWinClose()
  UnBufferVideo()
  Set(Cursor        , old_cursor)
//Set(SpecialEffects, old_SpecialEffects)

  if Query(Key) <> <Escape>
    attr_value = attr_select_value
  endif
end select_color_attribute


proc save_then_brighten_rgb_colors()
  integer brighten_percentage = 0
  integer b_color             = 0
  integer color_index         = 0
  integer ground              = 0
  integer g_color             = 0
  integer rgb_color           = 0
  integer r_color             = 0
  saved_rgb_colors = ''
  for ground = _FOREGROUND_ to _BACKGROUND_
    brighten_percentage = iif(ground == _FOREGROUND_,
                              foreground_percentage, background_percentage)
    for color_index = 0 to 15
      rgb_color        = GetColorTableValue(ground, color_index)
      saved_rgb_colors = Trim(saved_rgb_colors + ' ' + Str(rgb_color, 16))
      r_color = rgb_color / 65536
      g_color = (rgb_color mod 65536) / 256
      b_color = rgb_color mod 256
      if     brighten_percentage > 0
        r_color = (255 - r_color) * brighten_percentage / 100 + r_color
        g_color = (255 - g_color) * brighten_percentage / 100 + g_color
        b_color = (255 - b_color) * brighten_percentage / 100 + b_color
      elseif brighten_percentage < 0
        r_color = r_color - r_color * Abs(brighten_percentage) / 100
        g_color = g_color - g_color * Abs(brighten_percentage) / 100
        b_color = b_color - b_color * Abs(brighten_percentage) / 100
      endif
      rgb_color = (r_color * 65536) + (g_color * 256) + b_color
      SetColorTableValue(ground, color_index, rgb_color)
    endfor
  endfor
end save_then_brighten_rgb_colors


proc restore_rgb_colors()
  integer color_index   = 0
  integer ground        = 0
  string  rgb_color [6] = ''
  integer token_nr      = 0
  for ground = _FOREGROUND_ to _BACKGROUND_
    for color_index = 0 to 15
      token_nr  = token_nr + 1
      rgb_color = GetToken(saved_rgb_colors, ' ', token_nr)
      SetColorTableValue(ground, color_index, Val(rgb_color, 16))
    endfor
  endfor
end restore_rgb_colors


proc dim_brighten_cursor_line()
  integer old_cursor               = 0
  string  row_attrs [MAXSTRINGLEN] = ''
  string  row_chars [MAXSTRINGLEN] = ''
  integer screen_row_nr            = 0

  if Query(CursorAttr)        <> Query(TextAttr)
  or Query(CursorInBlockAttr) <> Query(BlockAttr)
    // Outside this extension one of the cursor attrs or a who color scheme
    // was changed.
    // Since we are still in dim/brighten mode, we save the new cursor attrs
    // for future use, and reset them to their non-cursor values.
    write_profile_int('TextAttr'         , Query(TextAttr))
    write_profile_int('BlockAttr'        , Query(BlockAttr))
    write_profile_int('CursorAttr'       , Query(CursorAttr))
    write_profile_int('CursorInBlockAttr', Query(CursorInBlockAttr))
    Set(CursorAttr       , Query(TextAttr))
    Set(CursorInBlockAttr, Query(BlockAttr))
  endif

  old_cursor    = Set(Cursor, OFF)
  screen_row_nr = Query(WindowY1) - 1 + CurrRow()

  // Brighten the current line
  GetStrAttrXY(Query(WindowX1), screen_row_nr, row_chars, row_attrs,
               Query(WindowCols))
  save_then_brighten_rgb_colors()
  Set(Attr, BLACK_ON_BLACK) // Works around the PutStrAttrXY() bug.
  PutStrAttrXY(Query(WindowX1), screen_row_nr, row_chars, '', BLACK_ON_BLACK)
  PutStrAttrXY(Query(WindowX1), screen_row_nr, row_chars, row_attrs)
  restore_rgb_colors()

  if  brightened_window_row_nr <> CurrRow()
  and brightened_file_line_nr  <> CurrLine()
  and brightened_buffer_id     == GetBufferId()
    // Restore the previously brightened line to not-brightened colors.
    GetStrAttrXY(Query(WindowX1), brightened_screen_row_nr, row_chars,
                 row_attrs, Query(WindowCols))
    Set(Attr, BLACK_ON_BLACK) // Works around the PutStrAttrXY() bug.
    PutStrAttrXY(Query(WindowX1), brightened_screen_row_nr, row_chars, '',
                 BLACK_ON_BLACK)
    PutStrAttrXY(Query(WindowX1), brightened_screen_row_nr, row_chars,
                 row_attrs)
  endif

  brightened_screen_row_nr = screen_row_nr
  brightened_window_row_nr = CurrRow()
  brightened_file_line_nr  = CurrLine()
  brightened_buffer_id     = GetBufferId()

  Set(Cursor, old_cursor)

end dim_brighten_cursor_line


proc idle_dims_brightens()
  if  update_cursor_line == -1
  and DisplayMode()      == _DISPLAY_HELP_
    update_cursor_line = FALSE
  elseif update_cursor_line
    dim_brighten_cursor_line()
    update_cursor_line = FALSE
  endif
end idle_dims_brightens


proc idle_quits()
  DelAutoLoadMacro(MACRO_NAME)
  PurgeMacro(CurrMacroFilename())
end idle_quits


proc after_command()
  if DisplayMode() <> _DISPLAY_HELP_
    dim_brighten_cursor_line()
    update_cursor_line = TRUE
  endif
end after_command


proc make_alarm_sound()
  if Query(Beep)
    Alarm()
  endif
end make_alarm_sound


proc make_sound(integer frequency, integer duration)
  if Query(Beep)
    Sound(frequency, duration)
  endif
end make_sound


integer proc ReadNum(integer n)
  // The string length being 4 is specific to this macro.
  // It allows percentages from "-100" to "100".
  string s[4] = Str(n)
  return (iif(ReadNumeric(s), Val(s), n))
end ReadNum


proc set_var(var integer var_ref)
  integer old_value = var_ref

  UnHook(after_getkey)

  var_ref = ReadNum(var_ref)

  if not (var_ref in -100 .. 100)
    Warn('The entered value'; var_ref; 'was not between -100 and 100.')
    var_ref = old_value
  endif

  Hook(_AFTER_GETKEY_, after_getkey)
end set_var


integer proc get_var(var integer var_ref)
  return(var_ref)
end get_var


proc after_getkey()
  integer adjust_for_mouse_click    = FALSE
  integer adjust_for_switching_line = FALSE
  integer direction                 = _NONE_
  integer key_code                  = Query(Key)
  integer mouse_x                   = Query(MouseX)
  integer mouse_y                   = Query(MouseY)
  integer new_value                 = 0
  integer old_value                 = 0
  integer where_y                   = WhereY()


  if   key_code == <LeftBtn>
  and  mouse_y  == 1
  and (mouse_x  in Query(PopWinCols) - 3 .. Query(PopWinCols) + 1)
    // We captured a mouse click on the menu close button we added.
    // Horizontally we accept clicked neighboring columns too.
    PushKey(<Escape>)
  else
    case color_type
      when 2
        if (key_code in  <b>, <B>, <c>, <C>)
        or (   (where_y  in CURSOR_LINE_ROW, CURSOR_LINE_IN_BLOCK_ROW)
           and (key_code in <Enter>, <GreyEnter>))
          PushKey(<Enter>)
          PushKey(<CursorDown>)
        endif
      when 3
        case key_code
          when <CursorLeft>,
               <GreyCursorLeft>
            direction = _LEFT_
          when <CursorRight>,
               <GreyCursorRight>
            direction = _RIGHT_
          when <LeftBtn>
            if   mouse_y == 1
            and (mouse_x in Query(PopWinCols) - 3 .. Query(PopWinCols) + 1)
              // We captured a mouse click on the menu close button we added.
              // Horizontally we accept clicked neighboring columns too.
              PushKey(<Escape>)
            elseif (mouse_x in 6 .. 8, 19 .. 21)
            and    (mouse_y in 7, 8)
              // We captured a mouse click on one of the 4 arrows.
              // Horizontally we accept clicked neighboring columns too.

              // The user might have clicked an arrow while the selected menu
              // line is still another one.
              adjust_for_mouse_click = TRUE
              // WhereY is the vertical coordinate of the current video output window.
              // MouseY is the clicked row in the entire display.
              if where_y <> mouse_y - 1
                where_y                   = mouse_y - 1
                adjust_for_switching_line = TRUE
              endif
              // Finally, we can process the individual arrow "buttons".
              if     (mouse_x in  6 ..  8) // Allow some horizontal inprecision.
                direction = _LEFT_
              elseif (mouse_x in 19 .. 21) // Allow some horizontal inprecision.
                direction = _RIGHT_
              endif
            endif
        endcase

        if direction <> _NONE_
          if (where_y in BACKGROUND_MENU_ROW, FOREGROUND_MENU_ROW)
            if where_y == BACKGROUND_MENU_ROW
              old_value = get_var(background_percentage)
            else
              old_value = get_var(foreground_percentage)
            endif
            if direction == _LEFT_
              new_value = Max(old_value - 1, -100)
            else
              new_value = Min(old_value + 1, 100)
            endif
            if new_value == old_value
              // Sound() needed because Alarm() is silenced while a key is being pressed.
              make_sound(1000, 200)
              // Testing says a Delay() is often needed to not silence Alarm().
              Delay(9)
              make_alarm_sound()
            else
              PushKey(<Enter>)
              if old_value == -100
                PushKey(<Del>)
              endif
              PushKeyStr(Str(new_value))
              PushKey(<Enter>)
              //  If an arrow was clicked while the current menu line was not the
              //  arrow's line, then we now adjust the menu's current line.
              //  Because PushKey()s are processed in reverse order,
              //  we have to do that last.
              if adjust_for_switching_line
                if where_y == FOREGROUND_MENU_ROW
                  PushKey(<CursorDown>)
                endif
                PushKey(<CursorDown>)
                PushKey(<CursorDown>)
                PushKey(<CursorDown>)
                PushKey(<Home>)
              endif
              if adjust_for_mouse_click
                PushKey(<Escape>)
              endif
            endif
          endif
        endif
    endcase
  endif
end after_getkey


string proc get_tse_color(integer color_attr_id)
  string  color_name [33] = ''
  integer color_attr      = 0

  if     color_attr_id == CURSOR_ATTR
    color_attr = Query(CursorAttr)
  elseif color_attr_id == CURSOR_IN_BLOCK_ATTR
    color_attr = Query(CursorInBlockAttr)
  endif

  if color_attr & 10000000b
    color_name = 'Intense'
  endif

  if color_attr & 00001000b
    color_name = Trim(color_name + ' Bright')
  endif

  color_name = Trim(color_name + ' '    +
                    GetToken(COLOR_NAMES, ' ', (color_attr & 00000111b) + 1))
  color_name = color_name + ' on ' +
               GetToken(COLOR_NAMES, ' ', (color_attr & 01110000b) / 16 + 1)
  return(color_name)
end get_tse_color


proc set_tse_color(integer color_attr_id)
  integer color_attr_value = 0

  if     color_attr_id == CURSOR_ATTR
    color_attr_value = Query(CursorAttr)
    select_color_attribute(color_attr_value)
    Set(CursorAttr, color_attr_value)
  elseif color_attr_id == CURSOR_IN_BLOCK_ATTR
    color_attr_value = Query(CursorInBlockAttr)
    select_color_attribute(color_attr_value)
    Set(CursorInBlockAttr, color_attr_value)
  endif
end set_tse_color


proc set_color_type(integer new_color_type)
  color_type = new_color_type
end set_color_type


integer proc get_color_type_checkmark_flag(integer menu_option_color_type)
  return(iif(color_type == menu_option_color_type,
             _MF_CHECKED_,
             _MF_UNCHECKED_))
end get_color_type_checkmark_flag


menu no_colortype_menu()
  title   = 'Cursor Line'
  x       = 1
  y       = 1
  history = main_menu_history
  CheckBoxes

  '&Same as other lines.',
    set_color_type(1),
    get_color_type_checkmark_flag(1)|_MF_CLOSE_ALL_BEFORE_,
    "Make the cursor line use the same colors as other lines."

  '&TSE color configuration.',
    set_color_type(2),
    get_color_type_checkmark_flag(2)|_MF_CLOSE_ALL_BEFORE_,
    "Make the cursor line use TSE's configuration for Editing colors."

  '&Dim/brighten colors.',
    set_color_type(3),
    get_color_type_checkmark_flag(3)|_MF_CLOSE_ALL_BEFORE_,
    "Make the cursor line use dimmed or brightened text colors."
end no_colortype_menu


menu tse_colortype_menu()
  title   = 'Cursor Line'
  x       = 1
  y       = 1
  history = main_menu_history
  CheckBoxes

  '&Same as other lines.',
    set_color_type(1),
    get_color_type_checkmark_flag(1)|_MF_CLOSE_ALL_BEFORE_,
    "Make the cursor line use the same colors as other lines."

  '&TSE color configuration',
    set_color_type(2),
    get_color_type_checkmark_flag(2)|_MF_CLOSE_ALL_BEFORE_,
    "Make the cursor line use TSE's configuration for Editing colors."

  '&Dim/brighten colors.',
    set_color_type(3),
    get_color_type_checkmark_flag(3)|_MF_CLOSE_ALL_BEFORE_,
    "Make the cursor line use dimmed or brightened text colors."

  '',,_MF_DIVIDE_

  '&Cursor Line',
    ,
    _MF_DONT_CLOSE_,
    "TSE's own cursor line color"

  ''
    [get_tse_color(CURSOR_ATTR):33],
    set_tse_color(CURSOR_ATTR),
    _MF_DONT_CLOSE_,
    "TSE's own cursor line color"

  'Cursor Line in &Block',
     ,
    _MF_DONT_CLOSE_,
    "TSE's own cursor line in block color"

  ''
    [get_tse_color(CURSOR_IN_BLOCK_ATTR):33],
    set_tse_color(CURSOR_IN_BLOCK_ATTR),
    _MF_DONT_CLOSE_,
    "TSE's own cursor line in block color"
end tse_colortype_menu


menu brighten_colortype_menu()
  title   = 'Cursor Line'
  x       = 1
  y       = 1
  history = main_menu_history
  CheckBoxes

  '&Same as other lines.',
    set_color_type(1),
    get_color_type_checkmark_flag(1)|_MF_CLOSE_ALL_BEFORE_,
    "Make cursor line colors the same as other lines."

  '&TSE color configuration.',
    set_color_type(2),
    get_color_type_checkmark_flag(2)|_MF_CLOSE_ALL_BEFORE_,
    "Make the cursor line use TSE's configuration for Editing colors."

  '&Dim/brighten colors',
    set_color_type(3),
    get_color_type_checkmark_flag(3)|_MF_CLOSE_ALL_BEFORE_,
    "Make the cursor line use dimmed or brightened text colors."

  '',,_MF_DIVIDE_

  'with a percentage:',,_MF_SKIP_

  '  ' + Chr(17) + ' &Background ' + Chr(16)
    [get_var(background_percentage):4],
    set_var(background_percentage),
    _MF_DONT_CLOSE_,
    'Use horizontal cursor keys or <Enter>.  -100% = black, 0% = same, +100% = white.'

  '  ' + Chr(17) + ' &Foreground ' + Chr(16)
    [get_var(foreground_percentage):4],
    set_var(foreground_percentage),
    _MF_DONT_CLOSE_,
    'Use horizontal cursor keys or <Enter>.  -100% = black, 0% = same, +100% = white.'
end brighten_colortype_menu


forward proc start_adding_menu_close_button()


proc stop_adding_menu_close_button()
  UnHook(start_adding_menu_close_button)
  UnHook(stop_adding_menu_close_button)
end stop_adding_menu_close_button


proc start_adding_menu_close_button()
  PutStrAttrXY(Query(PopWinCols) - 2, 0, '[X]', '', Query(MenuBorderAttr))

  //  Setting Transparency() is a trick to cause a screen update from inside
  //  a _NONEDIT_IDLE_ hook.
  //  The Transparency variable exists since TSE 4.4.
  //  For TSE 4.2 this means that the close button will initially not display,
  //  but only appear after clicking anywhere in the menu, even on a border.
  #if EDITOR_VERSION >= 4400h
    Set(Transparency, Query(Transparency))
  #endif

  stop_adding_menu_close_button()
end start_adding_menu_close_button


proc add_menu_close_button()
  Hook(_IDLE_        , stop_adding_menu_close_button)
  Hook(_NONEDIT_IDLE_, start_adding_menu_close_button)
end add_menu_close_button


proc configure()
  integer old_CursorAttr        = Query(CursorAttr)
  integer old_CursorInBlockAttr = Query(CursorInBlockAttr)
  integer old_SpecialEffects    = Query(SpecialEffects)

  Set(SpecialEffects, Query(SpecialEffects) & NO_POSITIONS_NO_SHADOWS)

  UnHook(idle_quits)
  UnHook(after_command)
  UnHook(idle_dims_brightens)
  Hook(_AFTER_GETKEY_, after_getkey)

  PopWinOpen(1, 1, Query(ScreenCols), Query(ScreenRows), 0, '',
             Query(MenuBorderAttr))
  Set(Attr, Query(TextAttr))
  ClrScr()
  main_menu_history = color_type

  if Query(TextAttr)  <> get_profile_int('TextAttr')
  or Query(BlockAttr) <> get_profile_int('BlockAttr')
    // The gemeral text or block color was changed outside this extension.
    // This is incompatible with this extension's stored settings.
    // The extension therefore resets its TSE color configuration.
    write_profile_int('TextAttr'         , Query(TextAttr))
    write_profile_int('BlockAttr'       , Query(BlockAttr))
    write_profile_int('CursorAttr'       , Query(CursorAttr))
    write_profile_int('CursorInBlockAttr', Query(CursorInBlockAttr))
  endif

  repeat
    case color_type
      when 1
        add_menu_close_button()
        no_colortype_menu()
      when 2
        if  Query(CursorAttr)        == Query(TextAttr)
        and Query(CursorInBlockAttr) == Query(BlockAttr)
          if Query(CursorAttr)        <> get_profile_int('CursorAttr')
          or Query(CursorInBlockAttr) <> get_profile_int('CursorInBlockAttr')
            Set(CursorAttr       , get_profile_int('CursorAttr'))
            Set(CursorInBlockAttr, get_profile_int('CursorInBlockAttr'))
          endif
        endif
        add_menu_close_button()
        tse_colortype_menu()
      when 3
        add_menu_close_button()
        brighten_colortype_menu()
      otherwise
        Warn(MACRO_NAME; 'program error: Illegal color type.')
        PushKey(<Escape>)
    endcase
    main_menu_history = MenuOption()
  until Query(Key) == <Escape>


  PopWinClose()
  Set(SpecialEffects , old_SpecialEffects )
  UnHook(after_getkey)

  write_profile_int('ColorType'           , color_type)
  write_profile_int('BackgroundPercentage', background_percentage)
  write_profile_int('ForegroundPercentage', foreground_percentage)

  if (color_type in 1, 3)
    if Query(CursorAttr)        <> Query(TextAttr)
    or Query(CursorInBlockAttr) <> Query(BlockAttr)
      write_profile_int('TextAttr'         , Query(TextAttr))
      write_profile_int('BlockAttr'        , Query(BlockAttr))
      write_profile_int('CursorAttr'       , Query(CursorAttr))
      write_profile_int('CursorInBlockAttr', Query(CursorInBlockAttr))
      Set(CursorAttr       , Query(TextAttr))
      Set(CursorInBlockAttr, Query(BlockAttr))
      SaveSettings()
    endif
  elseif color_type == 2
    if Query(CursorAttr)        <> get_profile_int('CursorAttr')
    or Query(CursorInBlockAttr) <> get_profile_int('CursorInBlockAttr')
      write_profile_int('TextAttr'         , Query(TextAttr))
      write_profile_int('BlockAttr'        , Query(BlockAttr))
      write_profile_int('CursorAttr'       , Query(CursorAttr))
      write_profile_int('CursorInBlockAttr', Query(CursorInBlockAttr))
    endif

    if Query(CursorAttr)        <> old_CursorAttr
    or Query(CursorInBlockAttr) <> old_CursorInBlockAttr
      SaveSettings()
    endif
  endif

  if (color_type in 1, 2)
    Hook(  _IDLE_         , idle_quits)
  elseif color_type == 3
    if  foreground_percentage == 0
    and background_percentage == 0
      Hook(_IDLE_         , idle_quits)
    else
      Hook(_AFTER_COMMAND_, after_command)
      Hook(_IDLE_         , idle_dims_brightens)

      if isAutoLoaded1('Uniview')
        // In Macro AutoLoad List "Uniview" must be higher than "CursorLine".
        // AddAutoLoadMacro() loads new entries on Macro AutoLoad List's top.
        DelAutoLoadMacro('Uniview')
        AddAutoLoadMacro(MACRO_NAME)
        AddAutoLoadMacro('Uniview')
      else
        AddAutoLoadMacro(MACRO_NAME)
      endif
    endif
  endif
end configure


proc Main()
  if tse_has_rgb_support()
    configure()
  else
    #ifdef LINUX
      Warn(MACRO_NAME;
          'error: You need to configure Linux TSE for RGB color support.',
          Chr(13), "Search for RGB in TSE's read.me.linux file.")
      PurgeMacro(MACRO_NAME)
    #endif
  endif
end Main


proc WhenPurged()
  if HASHES_ID
    AbandonFile(HASHES_ID)
  endif
end WhenPurged


proc WhenLoaded()
  integer org_id = GetBufferId()

  MACRO_NAME = SplitPath(CurrMacroFilename(), _NAME_)

  if tse_has_rgb_support()
    HASHES_ID = CreateTempBuffer()
    ChangeCurrFilename(MACRO_NAME + ':Hashes', CCF_OPTIONS)
    GotoBufferId(org_id)
    color_type            = get_profile_int('ColorType')
    background_percentage = get_profile_int('BackgroundPercentage')
    foreground_percentage = get_profile_int('ForegroundPercentage')
    if color_type == 0
      if  Query(CursorAttr)        == Query(TextAttr)
      and Query(CursorInBlockAttr) == Query(BlockAttr)
        color_type = 1
      else
        color_type = 2
      endif
    endif

    if (color_type in 1, 2)
      Hook(_IDLE_         , idle_quits)
    elseif color_type == 3
      Hook(_AFTER_COMMAND_, after_command)
      Hook(_IDLE_         , idle_dims_brightens)
    endif
  elseif tse_is_windows_console_version()
    if  GetProfileInt(MACRO_NAME + ':Config', 'RepeatConsoleWarning', TRUE)
    and MsgBox(MACRO_NAME + ' abort',
               Format(MACRO_NAME;
                      'requires the Windows GUI version of TSE.',
                      Chr(13), Chr(13),
                      'Repeat this error message in the future?'),
                _YES_NO_) == 2
      write_profile_int('RepeatConsoleWarning', FALSE)
    endif
    PurgeMacro(MACRO_NAME)
  endif
end WhenLoaded

